/*
 * Copyright (C) 2014 Canonical, Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: James Henstridge <james.henstridge@canonical.com>
 */

#include <algorithm>
#include <iostream>
#include <map>

#include <unity/scopes/Category.h>
#include <unity/scopes/CategorisedResult.h>
#include <unity/scopes/PreviewReply.h>
#include <unity/scopes/PreviewWidget.h>
#include <unity/scopes/Registry.h>
#include <unity/scopes/ScopeExceptions.h>
#include <unity/scopes/VariantBuilder.h>
#include "scopes-scope.h"
#include "resultforwarder.h"

using namespace unity::scopes;

static const char ONLINE_SCOPE_ID[] = "com.canonical.scopes.onlinescopes";

static const char MISSING_ICON[] = "/usr/share/icons/unity-icon-theme/places/svg/service-generic.svg";
static const char SCOPES_CATEGORY_DEFINITION[] = R"(
{
  "schema-version": 1,
  "template": {
    "category-layout": "grid",
    "card-size": "medium",
    "card-background": "color:///#E9E9E9"
  },
  "components": {
    "title": "title",
    "subtitle": "author",
    "mascot":  "icon",
    "background": "background"
  }
}
)";
// unconfuse emacs: "

int ScopesScope::start(std::string const&, RegistryProxy const &registry) {
    this->registry = registry;
    try {
        online_scope = registry->get_metadata(ONLINE_SCOPE_ID).proxy();
    } catch (std::exception &e) {
        std::cerr << "Could not instantiate online scopes scope: " << e.what() << std::endl;
    }
    return VERSION;
}

void ScopesScope::stop() {
    registry.reset();
}

QueryBase::UPtr ScopesScope::create_query(Query const &q,
                                          SearchMetadata const &hints) {
    QueryBase::UPtr query(new ScopesQuery(*this, q));
    return query;
}

QueryBase::UPtr ScopesScope::preview(Result const &result,
                                     ActionMetadata const &hints) {
    QueryBase::UPtr previewer(new ScopesPreview(*this, result));
    return previewer;
}

ActivationBase::UPtr ScopesScope::activate(Result const &result, ActionMetadata const& hints) {
    ActivationBase::UPtr activation(new ScopesActivation(result));
    return activation;
}

ActivationBase::UPtr ScopesScope::perform_action(Result const &result, ActionMetadata const& hints, std::string const &widget_id, std::string const &action_id) {
    ActivationBase::UPtr activation(new ScopesActivation(result));
    return activation;
}

ScopesQuery::ScopesQuery(ScopesScope &scope, Query const &query)
    : scope(scope), query(query) {
}

void ScopesQuery::cancelled() {
}

static std::string lowercase(std::string s) {
    std::transform(s.begin(), s.end(), s.begin(), ::tolower);
    return s;
}

enum ScopeCategory {
    CAT_FEATURED,
    CAT_ENTERTAINMENT,
    CAT_OTHER,
    N_CATEGORIES
};

static const std::map<std::string,ScopeCategory> category_mapping {
    {"com.canonical.scopes.amazon", CAT_FEATURED},
    {"com.canonical.scopes.ebay", CAT_FEATURED},
    {"com.canonical.scopes.grooveshark", CAT_ENTERTAINMENT},
    {"com.canonical.scopes.weatherchannel", CAT_FEATURED},
    {"com.canonical.scopes.wikipedia", CAT_FEATURED},
    {"musicaggregator", CAT_ENTERTAINMENT},
    {"videoaggregator", CAT_ENTERTAINMENT},
    {"clickscope", CAT_ENTERTAINMENT},
};

static const std::map<std::string,std::string> background_mapping {
    {"com.canonical.scopes.amazon", "color:///#FFFFFF"},
    {"com.canonical.scopes.ebay", "color:///#F4F4F4"},
    {"com.canonical.scopes.grooveshark", "color:///#F67F00"},
    {"com.canonical.scopes.weatherchannel", "color:///#244AA5"},
    {"com.canonical.scopes.wikipedia", "color:///#F6F6F6"},
};

static bool compareMetadata(ScopeMetadata const &item1, ScopeMetadata const &item2) {
    ScopeCategory cat1, cat2;
    try {
        cat1 = category_mapping.at(item1.scope_name());
    } catch (std::out_of_range &e) {
        cat1 = CAT_OTHER;
    }
    try {
        cat2 = category_mapping.at(item2.scope_name());
    } catch (std::out_of_range &e) {
        cat2 = CAT_OTHER;
    }

    // If categories differ, we're done.
    if (cat1 != cat2) {
        return cat1 < cat2;
    }

    std::string name1, name2;
    try {
        name1 = item1.display_name();
    } catch (NotFoundException &e) {
        name1 = item1.scope_name();
    }
    try {
        name2 = item2.display_name();
    } catch (NotFoundException &e) {
        name2 = item2.scope_name();
    }

    return name1 < name2;
}

void ScopesQuery::run(SearchReplyProxy const &reply) {
    CategoryRenderer renderer(SCOPES_CATEGORY_DEFINITION);
    Category::SCPtr categories[N_CATEGORIES];
    categories[CAT_FEATURED] = reply->register_category(
        "popular", "Featured", "", renderer);
    categories[CAT_ENTERTAINMENT] = reply->register_category(
        "entertainment", "Entertainment", "", renderer);
    categories[CAT_OTHER] = reply->register_category(
        "scopes", "Other", "", renderer);

    std::string term = lowercase(query.query_string());
    std::vector<ScopeMetadata> scopes;
    for (const auto &pair : scope.registry->list()) {
        const auto &item = pair.second;
        std::string display_name, description;
        try {
            display_name = lowercase(item.display_name());
        } catch (NotFoundException &e) {
        }
        try {
            description = lowercase(item.description());
        } catch (NotFoundException &e) {
        }
        if (display_name.find(term) != std::string::npos ||
            description.find(term) != std::string::npos) {
            scopes.push_back(item);
        }
    }
    std::sort(scopes.begin(), scopes.end(), compareMetadata);
    for (const auto &item : scopes) {
        // don't display ourselves
        if (item.scope_name() == "scopes") continue;

        std::string uri = "scope://" + item.scope_name();
        // TODO: categorisation of scopes should come from the
        // metadata rather than being hard coded.
        ScopeCategory cat_id;
        try {
            cat_id = category_mapping.at(item.scope_name());
        } catch (std::out_of_range &e) {
            cat_id = CAT_OTHER;
        }
        CategorisedResult result(categories[cat_id]);
        // Tell the shell to ask us to activate results.
        result.set_intercept_activation();
        result.set_uri(uri);
        try {
            result.set_title(item.display_name());
        } catch (NotFoundException &e) {
            result.set_title(item.scope_name());
        }
        try {
            result.set_art(item.art());
        } catch (NotFoundException &e) {
        }
        result.set_dnd_uri(uri);
        result["scope_name"] = item.scope_name();
        // FIXME: this should be provided by the scope metadata
        result["author"] = "Canonical Ltd";
        result["description"] = item.description();
        std::string icon;
        try {
            icon = item.icon();
        } catch (NotFoundException &e) {
        }
        if (icon.empty()) {
            result["icon"] = MISSING_ICON;
        } else {
            result["icon"] = icon;
        }
        try {
            result["background"] = background_mapping.at(item.scope_name());
        } catch (std::out_of_range &e) {
        }
        reply->push(result);
    }

    if (scope.online_scope && !query.query_string().empty()) {
        auto online_reply = std::make_shared<ResultForwarder>(reply);
        create_subquery(scope.online_scope, query.query_string(), online_reply);
    }
}

ScopesPreview::ScopesPreview(ScopesScope &scope, Result const &result)
    : scope(scope), result(result) {
}

void ScopesPreview::cancelled() {
}

void ScopesPreview::run(PreviewReplyProxy const &reply) {
    ColumnLayout layout1col(1), layout2col(2), layout3col(3);
    layout1col.add_column({"art", "header", "actions", "description"});

    layout2col.add_column({"art"});
    layout2col.add_column({"header", "actions", "description"});

    layout3col.add_column({"art"});
    layout3col.add_column({"header", "actions", "description"});
    layout3col.add_column({});
    reply->register_layout({layout1col, layout2col, layout3col});

    PreviewWidget header("header", "header");
    header.add_component("title", "title");
    header.add_component("subtitle", "author");
    header.add_component("mascot", "icon");

    PreviewWidget art("art", "image");
    art.add_component("source", "art");

    PreviewWidget description("description", "text");
    description.add_component("text", "description");

    PreviewWidget actions("actions", "actions");
    {
        VariantBuilder builder;
        builder.add_tuple({
                {"id", Variant("search")},
                {"label", Variant("Search")}
            });
        actions.add_attribute("actions", builder.end());
    }

    reply->push({art, header, actions, description});
}

ScopesActivation::ScopesActivation(Result const &result)
    : result(result) {
}

ActivationResponse ScopesActivation::activate() {
    Query query(result["scope_name"].get_string(), "", "");
    return ActivationResponse(query);
}

extern "C" ScopeBase *UNITY_SCOPE_CREATE_FUNCTION() {
    return new ScopesScope;
}

extern "C" void UNITY_SCOPE_DESTROY_FUNCTION(ScopeBase *scope) {
    delete scope;
}
